Attribute VB_Name = "modComplex"
'=======================================================
' Descrizione.....: Routines per le operazioni su
'                   numeri complessi.
' Nome dei Files..: Complex.bas
' Data............: 15/2/2000
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
'=======================================================
'   Comprende le seguenti funzioni e subroutines:
'
'   w = CCmp(r, i):     costruisce un numero complesso con
'                       parti reale e immaginaria date.
'
'   w = CCon(z):        ritorna il coniugato di z.
'   w = CSom(z1, z2):   ritorna la somma z1 + z2.
'   w = CDif(z1, z2):   ritorna la differenza z1 - z2.
'   w = CMol(z1, z2):   ritorna il prodotto z1 * z2.
'   w = CDiv(z1, z2):   ritorna il rapporto z1 / z2.
'
'   w = CPtN(z1, N):    ritorna la potenza z1 ^ N.
'   w = CPtc(z1, z2):   ritorna la potenza z1 ^ z2.
'   w = CSqr(z):        ritorna la radice quadrata di z.
'   CNnr z, N, w():     calcola in w() le N radici Nesime di z.
'   w = CExp(z):        ritorna l' esponenziale di z.
'   w = CLog(z):        ritorna il logaritmo naturale di z.
'
'   w = CSin(z):        ritorna il seno di z.
'   w = CCos(z):        ritorna il coseno di z.
'   w = CTan(z):        ritorna la tangente di z.
'   w = CAsin(z):       ritorna l' arcoseno di z.
'   w = CAcos(z):       ritorna l' arcocoseno di z.
'   w = CAtn(z):        ritorna l' arcotangente di z.
'   w = CSinh(z):       ritorna il seno iperbolico di z.
'   w = CCosh(z):       ritorna il coseno iperbolico di z.
'   w = CTanh(z):       ritorna la tangente iperbolica di z.
'   w = CAsinh(z):      ritorna l' arcoseno iperbolico di z.
'   w = CAcosh(z):      ritorna l' arcocoseno iperbolico di z.
'   w = CAtnh(z):       ritorna l' arcotangente iperbolica di z.
'
'   r = CAbs(z):        ritorna il valore assoluto di z.
'   r = CArg(z):        ritorna l' argomento di z.
'   r = CRea(z):        ritorna la parte reale di z.
'   i = CImm(z):        ritorna la parte immaginaria di z.
'   B = CEgu(z1, z2):   verifica l' eguaglianza z1 = z2.
'
'   S = CFormat$(z):    ritorna una stringa rappresentante z.
'   CMsgBox z:          espone un MsgBox con z a piena precisione.
'
'   r = Atn4Q(y, x):    ritorna il valore dell' ArcoTangente
'                       di y/x su 4 Quadranti.
'
'   Tipi:
'    i as Double, r as Double, x as Double, y as Double
'    w as Complex, z as Complex, z1 as Complex, z2 as Complex
'    N as Long
'    S as String
'    B as Boolean
'
Option Explicit
'
Public Type Complex
    re As Double  ' parte reale.
    im As Double  ' parte immaginaria.
End Type
'
Private Const PI# = 3.14159265358979    ' 4# * Atn(1#)
Private Const SMALNO# = 1E-99

Private Function Atn4Q(ByVal Y#, ByVal X#) As Double
'
'   Ritorna il valore dell' ArcoTangente di y/x su 4 Quadranti.
'   E':    -PI < Atn4Q <= PI.
'
    Select Case X
        Case Is > 0#
        Atn4Q = Atn(Y / X)
'
        Case Is < 0#
        If Y >= 0# Then
            Atn4Q = Atn(Y / X) + PI
        ElseIf Y < 0# Then
            Atn4Q = Atn(Y / X) - PI
        End If
'
        Case Is = 0#
        Atn4Q = CDbl(Sgn(Y)) * PI / 2#
    End Select
'
'
'
End Function
Public Function CArg(z As Complex) As Double
'
'   Ritorna l' argomento del numero complesso z:
'
    CArg = Atn4Q(z.im, z.re)
'
'
'
End Function
Public Function CAbs(z As Complex) As Double
'
'   Ritorna il valore assoluto (modulo) del numero complesso z:
'
    CAbs = Sqr(z.re * z.re + z.im * z.im)
'
'
'
End Function
Public Sub CMsgBox(z As Complex, Optional ByVal Txt$ = "")
'
'
    MsgBox CFormat$(z) & Txt$
'
'
'
End Sub

Public Function CPtN(z As Complex, ByVal N&) As Complex
'
'   Ritorna il valore, complesso, di z elevato alla N,
'   con N reale e intero:
'
    Dim J&, v As Complex
'
    v = CCmp(1#, 0#)
    For J = 1 To Abs(N)
        v = CMol(v, z)
    Next J
'
    If N < 0 Then
        CPtN = CDiv(CCmp(1#, 0#), v)
    Else
        CPtN = v
    End If
'
'
'
End Function
Public Function CDif(z1 As Complex, z2 As Complex) As Complex
'
'   Ritorna la differenza, complessa, dei due numeri complessi z1 - z2:
'
    CDif.re = z1.re - z2.re
    CDif.im = z1.im - z2.im
'
'
'
End Function
Public Function CSom(z1 As Complex, z2 As Complex) As Complex
'
'   Ritorna la somma, complessa, dei due numeri complessi z1 + z2:
'
    CSom.re = z1.re + z2.re
    CSom.im = z1.im + z2.im
'
'
'
End Function
Public Function CCmp(ByVal re#, ByVal im#) As Complex
'
'   Ritorna un numero complesso con parte reale re
'   e parte immaginaria im:
'
    CCmp.re = re
    CCmp.im = im
'
'
'
End Function
Public Function CMol(z1 As Complex, z2 As Complex) As Complex
'
'   Ritorna il prodotto, complesso, dei due numeri complessi z1 * z2:
'
    CMol.re = z1.re * z2.re - z1.im * z2.im
    CMol.im = z1.re * z2.im + z1.im * z2.re
'
'
'
End Function

Public Function CDiv(z1 As Complex, z2 As Complex _
    , Optional Errore As Boolean) As Complex
'
'   Ritorna il rapporto, complesso, dei due numeri complessi z1 / z2:
'
    Dim d#
'
    On Error GoTo CDiv_ERR
'
    d = z2.re * z2.re + z2.im * z2.im
'
    CDiv.re = (z1.re * z2.re + z1.im * z2.im) / d
    CDiv.im = (z1.im * z2.re - z1.re * z2.im) / d
'
'
CDiv_ERR:
    Errore = (Err <> 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, " Funzione CDiv"
        CDiv.re = CAbs(z1) * (Cos(CArg(z1) - CArg(z2))) / SMALNO
        CDiv.im = CAbs(z1) * (Sin(CArg(z1) - CArg(z2))) / SMALNO
    End If
'
'
'
End Function
Public Function CCon(z As Complex) As Complex
'
'   Ritorna il coniugato del numero complesso z:
'
    CCon.re = z.re
    CCon.im = -z.im
'
'
'
End Function


Public Function CExp(z As Complex _
    , Optional Errore As Boolean) As Complex
'
'   Ritorna il valore, complesso, dell' esponenziale di z:
'
    Dim ex#
'
    On Error GoTo CExp_ERR
'
    ex = Exp(z.re)
    CExp.re = ex * Cos(z.im)
    CExp.im = ex * Sin(z.im)
'
'
CExp_ERR:
    Errore = (Err <> 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, " Funzione CExp"
        ex = 1# / SMALNO
        CExp.re = ex * Cos(z.im)
        CExp.im = ex * Sin(z.im)
    End If
'
'
'
End Function

Public Function CEgu(z1 As Complex, z2 As Complex _
    , Optional d_re#, Optional d_im#) As Boolean
'
'   Verifica l' eguaglianza dei due numeri complessi z1 e z2.
'   Ritorna anche le differenze fra le parti reali e fra le
'   parti immaginarie:
'
    CEgu = (z1.re = z2.re) And (z1.im = z2.im)
    d_re = z1.re - z2.re
    d_im = z1.im - z2.im
'
'
'
End Function

Public Function CSqr(z As Complex) As Complex
'
'   Ritorna il valore principale, complesso,
'   della radice quadrata di z:
'
    Dim r_sq#, ar_2#
'
    r_sq = Sqr(CAbs(z))
    ar_2 = CArg(z) / 2#
'
    CSqr.re = r_sq * Cos(ar_2)
    CSqr.im = r_sq * Sin(ar_2)
'
'
'
End Function

Public Function CRea(z As Complex) As Double
'
'   Ritorna il valore della parte reale di z:
'
    CRea = z.re
'
'
'
End Function

Public Function CImm(z As Complex) As Double
'
'   Ritorna il valore della parte immaginaria di z:
'
    CImm = z.im
'
'
'
End Function

Public Function CFormat$(z As Complex _
    , Optional ByVal FS$ = "", Optional ByVal Nn& = 0)
'
'   Ritorna una stringa con la rappresentazione del
'   numero complesso z.
'   Parametri opzionali:
'    Fs$:   stringa di format. Segue la notazione standard
'           e viene applicata ugualmente alla parte reale e
'           a quella immaginaria. Se FS$ = "" o e' assente,
'           il numero viene rappresentato con la massima
'           precisione possibile.
'    Nn:    se > 0 vengono rappresentati solo gli Nn digits
'           piu' significativi. Se Nn non e' sufficiente a
'           rappresentare almeno la parte intera del numero,
'           viene prima tentata la notazione scientifica e
'           poi viene segnalato un errore.
'
    Dim Rea$, Imm$
'
    Rea$ = Trim$(Format$(Abs(z.re), FS$))
    Imm$ = Trim$(Format$(Abs(z.im), FS$))
'
    If Nn > 0 Then
        If InStr(Rea$, ".") - 1 <= Nn Then
            Rea$ = Left$(Rea$, Nn)
        ElseIf Nn >= 8 Then
            Rea$ = Format$(Abs(z.re), "0.00E+00")
        Else
            Rea$ = String$(Nn, "*")
        End If
'
        If InStr(Imm$, ".") - 1 <= Nn Then
            Imm$ = Left$(Imm$, Nn)
        ElseIf Nn >= 8 Then
            Imm$ = Format$(Abs(z.im), "0.00E+00")
        Else
            Imm$ = String$(Nn, "*")
        End If
    End If
'
    If z.re < 0# Then
        Rea$ = "-" & Rea$
    Else
        Rea$ = "+" & Rea$
    End If
'
    If z.im < 0# Then
        CFormat$ = Rea$ & " - " & Imm$ & "i"
    Else
        CFormat$ = Rea$ & " + " & Imm$ & "i"
    End If
'
'
'
End Function

Public Sub CNnr(z As Complex, ByVal N&, wRt() As Complex)
'
'   Calcola le N radici Nesime del numero complesso z
'   con N > 0.
'   Le radici vengono ritornate nel vettore wRt(1 To N):
'
    Dim J&, r#, ar#, ar_N#, ph_N#
    ReDim wRt(1 To N)
'
    r = (CAbs(z)) ^ (1# / CDbl(N))
    ar_N = CArg(z) / CDbl(N)
    ph_N = 2# * PI / CDbl(N)
'
    For J = 1 To N
        ar = ar_N + CDbl(J - 1) * ph_N
        wRt(J).re = r * Cos(ar)
        wRt(J).im = r * Sin(ar)
    Next J
'
'
'
End Sub

Public Function CLog(z As Complex _
    , Optional Errore As Boolean) As Complex
'
'   Ritorna il valore principale, complesso,
'   del logaritmo naturale di z:
'
    Dim ab#
'
    On Error GoTo CLog_ERR
'
    ab = CAbs(z)
'
    CLog.re = Log(ab)
    CLog.im = CArg(z)
'
'
CLog_ERR:
    Errore = (Err <> 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, " Funzione CLog"
        ab = SMALNO
        CLog.re = Log(ab)
        CLog.im = CArg(z)
    End If
'
'
'
End Function

Public Function CSin(z As Complex) As Complex
'
'   Ritorna il valore, complesso, del seno di z:
'
    Dim esp As Complex, esn As Complex
'
    esp = CMol(CCmp(0#, 1#), z)
    esn = CMol(CCmp(0#, -1#), z)
'
    CSin = CDiv(CDif(CExp(esp), CExp(esn)), CCmp(0#, 2#))
'
'
'
End Function

Public Function CCos(z As Complex) As Complex
'
'   Ritorna il valore, complesso, del coseno di z:
'
    Dim esp As Complex, esn As Complex
'
    esp = CMol(CCmp(0#, 1#), z)
    esn = CMol(CCmp(0#, -1#), z)
'
    CCos = CDiv(CSom(CExp(esp), CExp(esn)), CCmp(2#, 0#))
'
'
'
End Function

Public Function CTan(z As Complex) As Complex
'
'   Ritorna il valore, complesso, della tangente di z:
'
    CTan = CDiv(CSin(z), CCos(z))
'
'
'
End Function

Public Function CAtn(z As Complex) As Complex
'
'   Ritorna il valore, complesso, dell' arcotangente di z:
'
    Dim de As Complex, nu As Complex
'
    nu = CSom(CCmp(1#, 0#), CMol(CCmp(0#, 1), z))
    de = CDif(CCmp(1#, 0#), CMol(CCmp(0#, 1), z))
'
    CAtn = CDiv(CLog(CDiv(nu, de)), CCmp(0#, 2#))
'
'
'
End Function

Public Function CPtc(z As Complex, alfa As Complex) As Complex
'
'   Ritorna il valore principale, complesso, di z elevato ad alfa,
'   entrambi complessi:
'
    CPtc = CExp(CMol(alfa, CLog(z)))
'
'
'
End Function

Public Function CSinh(z As Complex) As Complex
'
'   Ritorna il valore, complesso, del seno
'   iperbolico di z:
'
    Dim esn As Complex
'
    esn = CMol(CCmp(-1#, 0#), z)
'
    CSinh = CDiv(CDif(CExp(z), CExp(esn)), CCmp(2#, 0#))
'
'
'
End Function

Public Function CCosh(z As Complex) As Complex
'
'   Ritorna il valore, complesso, del coseno
'   iperbolico di z:
'
    Dim esn As Complex
'
    esn = CMol(CCmp(-1#, 0#), z)
'
    CCosh = CDiv(CSom(CExp(z), CExp(esn)), CCmp(2#, 0#))
'
'
'
End Function

Public Function CTanh(z As Complex) As Complex
'
'   Ritorna il valore, complesso, della tangente
'   iperbolica di z:
'
    CTanh = CDiv(CSinh(z), CCosh(z))
'
'
'
End Function

Public Function CAtnh(z As Complex) As Complex
'
'   Ritorna il valore, complesso, dell' arcotangente
'   iperbolica di z:
'
    Dim de As Complex, nu As Complex
'
    nu = CSom(CCmp(1#, 0#), z)
    de = CDif(CCmp(1#, 0#), z)
'
    CAtnh = CDiv(CLog(CDiv(nu, de)), CCmp(2#, 0#))
'
'
'
End Function

Public Function CAsin(z As Complex) As Complex
'
'   Ritorna il valore, complesso, dell' arcoseno di z:
'
    Dim vc As Complex
'
    vc = CSqr(CDif(CCmp(1#, 0#), CMol(z, z)))
    vc = CSom(CMol(CCmp(0#, 1#), z), vc)
'
    CAsin = CDiv(CLog(vc), CCmp(0#, 1#))
'
'
'
End Function

Public Function CAcos(z As Complex) As Complex
'
'   Ritorna il valore, complesso, dell' arcocoseno di z:
'
    Dim vc As Complex
'
    vc = CSqr(CDif(CMol(z, z), CCmp(1#, 0#)))
    vc = CSom(z, vc)
'
    CAcos = CDiv(CLog(vc), CCmp(0#, 1#))
'
'
'
End Function

Public Function CAsinh(z As Complex) As Complex
'
'   Ritorna il valore, complesso, dell' arcoseno
'   iperbolico di z:
'
    Dim vc As Complex
'
    vc = CSqr(CSom(CMol(z, z), CCmp(1#, 0#)))
    vc = CSom(z, vc)
'
    CAsinh = CLog(vc)
'
'
'
End Function

Public Function CAcosh(z As Complex) As Complex
'
'   Ritorna il valore, complesso, dell' arcocoseno
'   iperbolico di z:
'
    Dim vc As Complex
'
    vc = CSqr(CDif(CMol(z, z), CCmp(1#, 0#)))
    vc = CSom(z, vc)
'
    CAcosh = CLog(vc)
'
'
'
End Function
